@@ -7,12 +7,14 @@
 from __future__ import absolute_import, print_function
 
 import base64
+import binascii
 import datetime
 import errno
 import importlib
 import io
 import json
 import logging
+import mimetypes
 import os
 import random
 import re
@@ -45,6 +47,7 @@ if version_info < (4,0):
 
 from tornado import httpserver
 from tornado import web
+from tornado.httputil import url_concat
 from tornado.log import LogFormatter, app_log, access_log, gen_log
 
 from notebook import (
@@ -52,7 +55,8 @@ from notebook import (
     DEFAULT_TEMPLATE_PATH_LIST,
     __version__,
 )
-from .base.handlers import Template404
+
+from .base.handlers import Template404, RedirectWithParams
 from .log import log_request
 from .services.kernels.kernelmanager import MappingKernelManager
 from .services.config import ConfigManager
@@ -236,15 +240,18 @@ class NotebookWebApplication(web.Application):
         # BEGIN HARDCODED WIDGETS HACK
         # TODO: Remove on notebook 5.0
         try:
-            import ipywidgets as widgets
-            handlers.append(
-                (r"/nbextensions/widgets/(.*)", FileFindHandler, {
-                    'path': widgets.find_static_assets(),
-                    'no_cache_paths': ['/'], # don't cache anything in nbextensions
-                }),
-            )
+            import widgetsnbextension
         except:
-            app_log.warning('Widgets are unavailable. Please install widgetsnbextension or ipywidgets 4.0')
+            try:
+                import ipywidgets as widgets
+                handlers.append(
+                    (r"/nbextensions/widgets/(.*)", FileFindHandler, {
+                        'path': widgets.find_static_assets(),
+                        'no_cache_paths': ['/'], # don't cache anything in nbextensions
+                    }),
+                )
+            except:
+                app_log.warning('Widgets are unavailable. Please install widgetsnbextension or ipywidgets 4.0')
         # END HARDCODED WIDGETS HACK
         
         handlers.append(
@@ -263,7 +270,7 @@ class NotebookWebApplication(web.Application):
         handlers.extend(load_handlers('base.handlers'))
         # set the URL that will be redirected from `/`
         handlers.append(
-            (r'/?', web.RedirectHandler, {
+            (r'/?', RedirectWithParams, {
                 'url' : settings['default_url'],
                 'permanent': False, # want 302, not 301
             })
@@ -300,7 +307,10 @@ class NbserverListApp(JupyterApp):
             if self.json:
                 print(json.dumps(serverinfo))
             else:
-                print(serverinfo['url'], "::", serverinfo['notebook_dir'])
+                url = serverinfo['url']
+                if serverinfo.get('token'):
+                    url = url + '?token=%s' % serverinfo['token']
+                print(url, "::", serverinfo['notebook_dir'])
 
 #-----------------------------------------------------------------------------
 # Aliases and Flags
@@ -467,7 +477,7 @@ class NotebookApp(JupyterApp):
     )
     
     client_ca = Unicode(u'', config=True,
-        help="""The full path to a certificate authority certifificate for SSL/TLS client authentication."""
+        help="""The full path to a certificate authority certificate for SSL/TLS client authentication."""
     )
     
     cookie_secret_file = Unicode(config=True,
@@ -507,6 +517,29 @@ class NotebookApp(JupyterApp):
                 self.cookie_secret_file
             )
 
+    token = Unicode(config=True,
+        help="""Token used for authenticating first-time connections to the server.
+
+        Only used when no password is enabled.
+
+        Setting to an empty string disables authentication altogether, which is NOT RECOMMENDED.
+        """
+    )
+
+    one_time_token = Unicode(
+        help="""One-time token used for opening a browser.
+
+        Once used, this token cannot be used again.
+        """
+    )
+
+    def _token_default(self):
+        if self.password:
+            # no token if password is enabled
+            return u''
+        else:
+            return binascii.hexlify(os.urandom(24)).decode('ascii')
+
     password = Unicode(u'', config=True,
                       help="""Hashed password to use for web authentication.
 
@@ -651,7 +684,10 @@ class NotebookApp(JupyterApp):
         """
     )
     mathjax_url = Unicode("", config=True,
-        help="""The url for MathJax.js."""
+        help="""A custom url for MathJax.js.
+        Should be in the form of a case-sensitive url to MathJax,
+        for example:  /static/components/MathJax/MathJax.js
+        """
     )
     def _mathjax_url_default(self):
         if not self.enable_mathjax:
@@ -788,7 +824,8 @@ class NotebookApp(JupyterApp):
     nbserver_extensions = Dict({}, config=True,
         help=("Dict of Python modules to load as notebook server extensions."
               "Entry values can be used to enable and disable the loading of"
-              "the extensions.")
+              "the extensions. The extensions will be loaded in alphabetical "
+              "order.")
     )
 
     reraise_server_extension_failures = Bool(
@@ -876,6 +913,11 @@ class NotebookApp(JupyterApp):
             self.tornado_settings['allow_origin_pat'] = re.compile(self.allow_origin_pat)
         self.tornado_settings['allow_credentials'] = self.allow_credentials
         self.tornado_settings['cookie_options'] = self.cookie_options
+        self.tornado_settings['token'] = self.token
+        if (self.open_browser or self.file_to_run) and not self.password:
+            self.one_time_token = binascii.hexlify(os.urandom(24)).decode('ascii')
+            self.tornado_settings['one_time_token'] = self.one_time_token
+
         # ensure default_url starts with base_url
         if not self.default_url.startswith(self.base_url):
             self.default_url = url_path_join(self.base_url, self.default_url)
@@ -900,10 +942,10 @@ class NotebookApp(JupyterApp):
         else:
             # SSL may be missing, so only import it if it's to be used
             import ssl
-            # Disable SSLv3, since its use is discouraged.
-            ssl_options['ssl_version'] = ssl.PROTOCOL_TLSv1
+            # Disable SSLv3 by default, since its use is discouraged.
+            ssl_options.setdefault('ssl_version', ssl.PROTOCOL_TLSv1)
             if ssl_options.get('ca_certs', False):
-                ssl_options['cert_reqs'] = ssl.CERT_REQUIRED
+                ssl_options.setdefault('cert_reqs', ssl.CERT_REQUIRED)
         
         self.login_handler_class.validate_security(self, ssl_options=ssl_options)
         self.http_server = httpserver.HTTPServer(self.web_app, ssl_options=ssl_options,
@@ -934,7 +976,8 @@ class NotebookApp(JupyterApp):
     @property
     def display_url(self):
         ip = self.ip if self.ip else '[all ip addresses on your system]'
-        return self._url(ip)
+        query = '?token=%s' % self.token if self.token else ''
+        return self._url(ip) + query
 
     @property
     def connection_url(self):
@@ -1036,7 +1079,7 @@ class NotebookApp(JupyterApp):
             if not modulename in self.nbserver_extensions:
                 self.nbserver_extensions[modulename] = True
         
-        for modulename in self.nbserver_extensions:
+        for modulename in sorted(self.nbserver_extensions):
             if self.nbserver_extensions[modulename]:
                 try:
                     mod = importlib.import_module(modulename)
@@ -1049,6 +1092,13 @@ class NotebookApp(JupyterApp):
                     self.log.warning("Error loading server extension %s", modulename,
                                   exc_info=True)
 
+    def init_mime_overrides(self):
+        # On some Windows machines, an application has registered an incorrect
+        # mimetype for CSS in the registry. Tornado uses this when serving
+        # .css files, causing browsers to reject the stylesheet. We know the
+        # mimetype always needs to be text/css, so we override it here.
+        mimetypes.add_type('text/css', '.css')
+
     @catch_config_error
     def initialize(self, argv=None):
         super(NotebookApp, self).initialize(argv)
@@ -1061,6 +1111,7 @@ class NotebookApp(JupyterApp):
         self.init_terminals()
         self.init_signal()
         self.init_server_extensions()
+        self.init_mime_overrides()
 
     def cleanup_kernels(self):
         """Shutdown all kernels.
@@ -1084,14 +1135,16 @@ class NotebookApp(JupyterApp):
                 'port': self.port,
                 'secure': bool(self.certfile),
                 'base_url': self.base_url,
+                'token': self.token,
                 'notebook_dir': os.path.abspath(self.notebook_dir),
-                'pid': os.getpid()
+                'password': bool(self.password),
+                'pid': os.getpid(),
                }
 
     def write_server_info_file(self):
         """Write the result of server_info() to the JSON file info_file."""
         with open(self.info_file, 'w') as f:
-            json.dump(self.server_info(), f, indent=2)
+            json.dump(self.server_info(), f, indent=2, sort_keys=True)
 
     def remove_server_info_file(self):
         """Remove the nbserver-<pid>.json file created for this server.
@@ -1135,6 +1188,8 @@ class NotebookApp(JupyterApp):
             else:
                 # default_url contains base_url, but so does connection_url
                 uri = self.default_url[len(self.base_url):]
+            if self.one_time_token:
+                uri = url_concat(uri, {'token': self.one_time_token})
             if browser:
                 b = lambda : browser.open(url_path_join(self.connection_url, uri),
                                           new=2)
@@ -1151,9 +1206,9 @@ class NotebookApp(JupyterApp):
         except KeyboardInterrupt:
             info("Interrupted...")
         finally:
-            self.cleanup_kernels()
             self.remove_server_info_file()
-    
+            self.cleanup_kernels()
+
     def stop(self):
         def _stop():
             self.http_server.stop()
@@ -1187,7 +1242,7 @@ def list_running_servers(runtime_dir=None):
             else:
                 # If the process has died, try to delete its info file
                 try:
-                    os.unlink(file)
+                    os.unlink(os.path.join(runtime_dir, file))
                 except OSError:
                     pass  # TODO: This should warn or log or something
 #-----------------------------------------------------------------------------
