@@ -13,6 +13,16 @@ define([
     // keep track of which extensions have been loaded already
     var extensions_loaded = [];
 
+    /**
+     * Whether or not an extension has been loaded
+     * @param  {string} extension - name of the extension
+     * @return {boolean}            true if loaded already
+     */
+    var is_loaded = function(extension) {
+        var ext_path = "nbextensions/" + extension;
+        return extensions_loaded.indexOf(ext_path) >= 0;
+    };
+
     /**
      * Load a single extension.
      * @param  {string} extension - extension path.
@@ -21,17 +31,17 @@ define([
     var load_extension = function (extension) {
         return new Promise(function(resolve, reject) {
             var ext_path = "nbextensions/" + extension;
-            require([ext_path], function(module) {
-                try {
-                    if (extensions_loaded.indexOf(ext_path) < 0) {
-                        console.log("Loading extension: " + extension);
-                        module.load_ipython_extension();
-                        extensions_loaded.push(ext_path);
+            requirejs([ext_path], function(module) {
+                if (!is_loaded(extension)) {
+                    console.log("Loading extension: " + extension);
+                    if (module.load_ipython_extension) {
+                        Promise.resolve(module.load_ipython_extension()).then(function() {
+                            resolve(module);
+                        }).catch(reject);
                     }
-                    else{
-                        console.log("Loaded extension already: " + extension);
-                    }
-                } finally {
+                    extensions_loaded.push(ext_path);
+                } else {
+                    console.log("Loaded extension already: " + extension);
                     resolve(module);
                 }
             }, function(err) {
@@ -46,23 +56,38 @@ define([
      * @return {Promise} that resolves to a list of loaded module handles.
      */
     var load_extensions = function () {
+        console.log("load_extensions", arguments);
         return Promise.all(Array.prototype.map.call(arguments, load_extension)).catch(function(err) {
             console.error("Failed to load extension" + (err.requireModules.length>1?'s':'') + ":", err.requireModules, err);
         });
     };
 
+    /**
+     * Return a list of extensions that should be active
+     * The config for nbextensions comes in as a dict where keys are 
+     * nbextensions paths and the values are a bool indicating if it 
+     * should be active. This returns a list of nbextension paths
+     * where the value is true
+     */
+    function filter_extensions(nbext_config) {
+        var active = [];
+        Object.keys(nbext_config).forEach(function (nbext) {
+            if (nbext_config[nbext]) {active.push(nbext);}
+        });
+        return active;
+    }
+
     /**
      * Wait for a config section to load, and then load the extensions specified
      * in a 'load_extensions' key inside it.
      */
     function load_extensions_from_config(section) {
-        section.loaded.then(function() {
+        return section.loaded.then(function() {
             if (section.data.load_extensions) {
-                var nbextension_paths = Object.getOwnPropertyNames(
-                                            section.data.load_extensions);
-                load_extensions.apply(this, nbextension_paths);
+                var active = filter_extensions(section.data.load_extensions);
+                return load_extensions.apply(this, active);
             }
-        });
+        }).catch(utils.reject('Could not load nbextensions from ' + section.section_name + ' config file'));
     }
 
     //============================================================================
@@ -359,12 +384,13 @@ define([
     // Remove chunks that should be overridden by the effect of
     // carriage return characters
     function fixCarriageReturn(txt) {
-        var tmp = txt;
-        do {
-            txt = tmp;
-            tmp = txt.replace(/\r+\n/gm, '\n'); // \r followed by \n --> newline
-            tmp = tmp.replace(/^.*\r+/gm, '');  // Other \r --> clear line
-        } while (tmp.length < txt.length);
+        txt = txt.replace(/\r+\n/gm, '\n'); // \r followed by \n --> newline
+        while (txt.search(/\r/g) > -1) {
+            var base = txt.match(/^.*\r+/m)[0].replace(/\r/, '');
+            var insert = txt.match(/\r+.*$/m)[0].replace(/\r/, '');
+            insert = insert + base.slice(insert.length, base.length);
+            txt = txt.replace(/\r+.*$/m, '\r').replace(/^.*\r+/m, insert);
+        }
         return txt;
     }
 
@@ -826,8 +852,10 @@ define([
     };
     
     var utils = {
+        is_loaded: is_loaded,
         load_extension: load_extension,
         load_extensions: load_extensions,
+        filter_extensions: filter_extensions,
         load_extensions_from_config: load_extensions_from_config,
         regex_split : regex_split,
         uuid : uuid,
